use crate::components::error::error_provider::ErrorContext;
use crate::services::rest::RestService;
use common::CreateMilestone;
use std::rc::Rc;
use wasm_bindgen::JsCast;
use wasm_bindgen_futures::spawn_local;
use web_sys::HtmlInputElement;
use yew::html;
use yew::prelude::*;
use yew::Callback;
use yew::Html;
use yew_router::prelude::use_navigator;

#[derive(Properties, PartialEq)]
pub struct Props;

#[function_component]
pub fn CreateMilestoneComponent(_: &Props) -> Html {
    let nav = use_navigator().expect("cannot get navigator");
    let err_ctx = Rc::new(use_context::<ErrorContext>());

    let input_ref = use_node_ref();
    // Focus input after it is rendered
    {
        let input_ref = input_ref.clone();
        use_effect(move || {
            if let Some(input_element) = input_ref.get() {
                let input_element = input_element
                    .dyn_into::<HtmlInputElement>()
                    .expect("Failed to cast input element");
                input_element.focus().expect("Failed to focus input");
            }

            // effect destructor
            || {}
        });
    }

    let input_value = use_state(String::default);
    let awaiting_response = use_state(|| false);

    let onclick_go_back = {
        let nav = nav.clone();
        Callback::from(move |_: web_sys::MouseEvent| {
            nav.push(&crate::Route::Admin);
        })
    };

    let onsubmit = {
        let input_value = input_value.clone();
        let awaiting_response = awaiting_response.clone();

        Callback::from(move |e: yew::SubmitEvent| {
            e.prevent_default();

            log::info!("Creating milestone");
            let Ok(goal) = input_value.parse::<f64>() else {
                log::error!("Input parse error");
                return;
            };

            let goal = goal as usize;
            let payload = CreateMilestone { goal };
            {
                let nav = nav.clone();
                let err_ctx = Rc::clone(&err_ctx);
                awaiting_response.set(true);
                let awaiting_response = awaiting_response.clone();
                spawn_local(async move {
                    match RestService::create_milestone(payload).await {
                        Ok(_response) => nav.push(&crate::Route::Admin),
                        Err(err) => {
                            if let Some(err_ctx) = &*err_ctx {
                                err_ctx.dispatch(err.to_string());
                            }
                        }
                    };
                    awaiting_response.set(false);
                });
            }
        })
    };

    let oninput = {
        let input_value = input_value.clone();

        Callback::from(move |e: web_sys::InputEvent| {
            let Some(input) = e
                    .target()
                    .and_then(|t| t.dyn_into::<web_sys::HtmlInputElement>().ok()) else { unreachable!() };
            input_value.set(input.value());
        })
    };

    html! {
        <>
        <div class="row">
            <button onclick={onclick_go_back} class="button">{"Back"}</button>
        </div>

        <form {onsubmit} >
          <hr />
          <div class="row">
            <div class="twelve columns">
              <label for="milestoneInput">{"New milestone"}</label>
              <input ref={input_ref.clone()} {oninput} value={input_value.to_string()} class="u-full-width" type="number" id="milestoneInput" />
            </div>
          </div>
          <input class="button-primary" type="submit" value="Submit" disabled={*awaiting_response} />
        </form>
        </>
    }
}
