use crate::event_bus::EventBus;
use crate::event_bus::EventBusInput;
use futures::channel::mpsc::Sender;
use futures::SinkExt;
use futures::StreamExt;
use reqwasm::websocket::futures::WebSocket;
use reqwasm::websocket::Message;
use std::time::Duration;
use wasm_bindgen_futures::spawn_local;
use yew_agent::Dispatched;

#[derive(Debug)]
pub struct WebsocketService {
    pub tx: Sender<common::WebSocketMessageAppToServer>,
}

impl WebsocketService {
    pub fn connect() -> Self {
        let window = web_sys::window().expect("no global `window` exists");
        let location = window.location();
        let protocol = location.protocol().expect("should have a protocol");
        let hostname = location.hostname().expect("should have a hostname");

        let ws_protocol = match protocol.as_str() {
            "https:" => "wss",
            _ => "ws",
        };

        let port = match location.port() {
            Ok(port) if !port.is_empty() => format!(":{}", port),
            _ => String::new(),
        };

        let ws_url = format!("{}://{}{}/api/ws", ws_protocol, hostname, port);
        let ws = WebSocket::open(&ws_url).unwrap();
        log::info!("Opened websocket connection to {ws_url}");

        let (write, mut read) = ws.split();

        let (in_tx, in_rx) = futures::channel::mpsc::channel::<common::WebSocketMessageAppToServer>(1000);
        let mut event_bus = EventBus::<common::State>::dispatcher();

        // Generate regular heartbeat messages from app to server
        let (mut heartbeat_tx, heartbeat_rx) = futures::channel::mpsc::channel(1);
        spawn_local(async move {
            let heartbeat_interval = yew::platform::time::interval(Duration::from_secs(20));
            futures::pin_mut!(heartbeat_interval);
            while heartbeat_interval.next().await.is_some() {
                let msg = common::WebSocketMessageAppToServer::HeartBeat;
                heartbeat_tx.send(msg).await.unwrap();
            }
        });

        // App to Server
        let all_tx = futures::stream::select(in_rx, heartbeat_rx);
        spawn_local(async move {
            // Serialize as JSON and map to websocket string message
            let all_tx = all_tx
                .map(|msg| serde_json::to_string(&msg).expect("Serialization error"))
                .map(Message::Text)
                .map(Ok);
            all_tx
                .forward(write)
                .await
                .expect("Forward to websocket write half failed");
        });

        // Server to App
        spawn_local(async move {
            while let Some(msg) = read.next().await {
                match msg {
                    Ok(Message::Text(data)) => {
                        match serde_json::from_str::<common::WebSocketMessageServerToApp>(&data) {
                            Ok(ws_msg) => match ws_msg {
                                common::WebSocketMessageServerToApp::State(state) => {
                                    log::debug!("Received ws message. Dispatching to event bus.");
                                    event_bus.send(EventBusInput::EventBusMsg(state));
                                }
                                common::WebSocketMessageServerToApp::HeartBeat => {}
                            },
                            Err(err) => {
                                log::error!("{err:?}");
                            }
                        }
                    }
                    Ok(Message::Bytes(_)) => {
                        log::warn!("Received binary data on websocket. This is unhandled");
                    }
                    Err(err) => {
                        log::error!("ws error: {err:?}");
                    }
                }
            }
            log::debug!("WebSocket Closed");
        });

        Self { tx: in_tx }
    }
}
