use chrono::NaiveTime;
use common::Achievement;

pub fn group_achievements(mut achievements: Vec<Achievement>) -> Vec<(Option<NaiveTime>, Vec<Achievement>)> {
    // Sort the achievements by time_of_reveal.
    achievements.sort_by_key(|a| a.time_of_reveal);

    let mut grouped_achievements: Vec<(Option<NaiveTime>, Vec<Achievement>)> = Vec::new();
    let mut current_time_of_reveal: Option<NaiveTime> = None;
    let mut current_group: Vec<Achievement> = Vec::new();

    for achievement in achievements {
        if current_time_of_reveal == achievement.time_of_reveal {
            current_group.push(achievement);
            continue;
        }
        if !current_group.is_empty() {
            grouped_achievements.push((current_time_of_reveal, current_group));
        }
        current_time_of_reveal = achievement.time_of_reveal;
        current_group = vec![achievement];
    }

    if !current_group.is_empty() {
        grouped_achievements.push((current_time_of_reveal, current_group));
    }

    grouped_achievements
}

#[cfg(test)]
mod tests {
    use super::*;
    use uuid::Uuid;

    #[test]
    fn test_group_achievements() {
        let time1 = NaiveTime::from_hms_opt(12, 0, 0).unwrap();
        let time2 = NaiveTime::from_hms_opt(13, 0, 0).unwrap();

        let achievements = vec![
            Achievement {
                goal: "Goal 1".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: Some(time1),
            },
            Achievement {
                goal: "Goal 2".to_string(),
                completed: false,
                uuid: Uuid::new_v4(),
                time_of_reveal: Some(time1),
            },
            Achievement {
                goal: "Goal 3".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: Some(time2),
            },
        ];

        let grouped = group_achievements(achievements);
        assert_eq!(grouped.len(), 2);
        assert_eq!(grouped[0].0, Some(time1));
        assert_eq!(grouped[0].1.len(), 2);
        assert_eq!(grouped[1].0, Some(time2));
        assert_eq!(grouped[1].1.len(), 1);
    }

    #[test]
    fn test_group_achievements_empty_vec() {
        let achievements = vec![];
        let grouped = group_achievements(achievements);
        assert_eq!(grouped.len(), 0);
    }

    #[test]
    fn test_group_achievements_group_none_values() {
        let time1 = NaiveTime::from_hms_opt(12, 0, 0).unwrap();

        let achievements = vec![
            Achievement {
                goal: "Goal 1".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: Some(time1),
            },
            Achievement {
                goal: "Goal 2".to_string(),
                completed: false,
                uuid: Uuid::new_v4(),
                time_of_reveal: None,
            },
            Achievement {
                goal: "Goal 3".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: None,
            },
        ];

        let grouped = group_achievements(achievements);
        assert_eq!(grouped.len(), 2);
        assert_eq!(grouped[0].0, None);
        assert_eq!(grouped[0].1.len(), 2);
        assert_eq!(grouped[1].0, Some(time1));
        assert_eq!(grouped[1].1.len(), 1);
    }

    #[test]
    fn test_group_achievements_all_none() {
        let achievements = vec![
            Achievement {
                goal: "Goal 1".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: None,
            },
            Achievement {
                goal: "Goal 2".to_string(),
                completed: false,
                uuid: Uuid::new_v4(),
                time_of_reveal: None,
            },
            Achievement {
                goal: "Goal 3".to_string(),
                completed: true,
                uuid: Uuid::new_v4(),
                time_of_reveal: None,
            },
        ];

        let grouped = group_achievements(achievements);
        assert_eq!(grouped.len(), 1);
        assert_eq!(grouped[0].0, None);
        assert_eq!(grouped[0].1.len(), 3);
    }
}
